#pylint: disable=logging-format-interpolation,logging-not-lazy,too-many-arguments,too-many-locals,too-many-statements
import base64
import logging
import traceback
from datetime import date
from os.path import basename
import requests
from django.conf import settings
from django.template.loader import get_template
from django.utils.timezone import now
from django.utils.translation import ugettext as _
from zeep.wsse import BinarySignature
from jbank.models import WsEdiConnection, WsEdiSoapCall
from lxml import etree


logger = logging.getLogger(__name__)


def wsedi_get(command: str, file_type: str, status: str, file_reference: str = '', verbose: bool = False) -> requests.Response:
    """
    Download Finnish bank files. Assumes WS-EDI API parameter compatible HTTP REST API end-point.
    Uses project settings WSEDI_URL and WSEDI_TOKEN.
    :param command: Command, e.g. DownloadFileList or DownloadFile
    :param file_type: File type, e.g. TO or SVM
    :param status: Status, e.g. DLD or NEW
    :param file_reference: File reference (if command is DownloadFile)
    :param verbose: Debug output
    :return: requests.Response
    """
    url = settings.WSEDI_URL + '?command={command}'.format(command=command)
    if file_reference:
        url += '&file-reference=' + file_reference
    if file_type:
        url += '&file-type=' + file_type
    if status:
        url += '&status=' + status
    headers = {
        'Content-Type': 'application/json',
        'Authorization': 'Token ' + settings.WSEDI_TOKEN,
    }
    res = requests.get(url, headers=headers)
    if res.status_code >= 300:
        logger.error("wsedi_get(command={}, file_type={}, status={}, file_reference={}) response HTTP {}:\n".format(
            command, file_type, status, file_reference, res.status_code) + res.text)
    elif verbose:
        logger.info("wsedi_get(command={}, file_type={}, status={}, file_reference={}) response HTTP {}:\n".format(
            command, file_type, status, file_reference, res.status_code) + res.text)

    if res.status_code >= 300:
        raise Exception("WS-EDI {} HTTP {}".format(command, res.status_code))
    return res


def wsedi_upload_file(file_content: str, file_type: str, file_name: str, verbose: bool = False) -> requests.Response:
    """
    Upload Finnish bank file. Assumes WS-EDI API parameter compatible HTTP REST API end-point.
    Uses project settings WSEDI_URL and WSEDI_TOKEN.
    :param file_content: File content
    :param file_type: File type, e.g. pain.001.001.03
    :param file_name: File (base) name
    :param verbose: Debug output
    :return: requests.Response
    """
    command = 'UploadFile'
    url = settings.WSEDI_URL
    data = {
        'command': command,
        'file-type': file_type,
        'file-name': basename(file_name),
        'file-content': base64.b64encode(file_content.encode('utf8')).decode('ascii'),
    }
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Token ' + settings.WSEDI_TOKEN,
    }
    res = requests.post(url, data=data, headers=headers)
    if res.status_code >= 300:
        logger.error("wsedi_upload_file(command={}, file_type={}, file_name={}) response HTTP {}:\n".format(
            command, file_type, file_name, res.status_code) + res.text)
        raise Exception("WS-EDI {} HTTP {}".format(command, res.status_code))
    if verbose:
        logger.info("wsedi_upload_file(command={}, file_type={}, file_name={}) response HTTP {}:\n".format(
            command, file_type, file_name, res.status_code) + res.text)
    return res


def wsedi_execute(ws: WsEdiConnection, command: str, file_type: str = '', status: str = '', file_reference: str = '',  # noqa
                  file_content: str = '', start_date: date or None = None, end_date: date or None = None,
                  verbose: bool = False, cls: callable = WsEdiSoapCall, **kwargs) -> bytes:
    """
    :param ws:
    :param command:
    :param file_type:
    :param status:
    :param file_reference:
    :param file_content:
    :param start_date:
    :param end_date:
    :param verbose:
    :param cls:
    :return: str
    """
    if ws and not ws.enabled:
        raise Exception(_('ws.edi.connection.not.enabled').format(ws=ws))

    soap_call = cls(connection=ws, command=command, **kwargs)
    soap_call.full_clean()
    soap_call.save()
    call_str = 'WsEdiSoapCall({})'.format(soap_call.id)
    try:
        content = ''
        if file_content:
            content = base64.b64encode(file_content.encode()).decode('ascii')

        app = ws.get_application_request(command, file_type=file_type, status=status, file_reference=file_reference, content=content,
                                         start_date=start_date, end_date=end_date)
        if verbose:
            logger.info('------------------------------------------------------ {} app\n{}'.format(call_str, app))
        debug_output = command in ws.debug_command_list or 'ALL' in ws.debug_command_list
        if debug_output:
            with open(soap_call.debug_application_request_full_path, 'wb') as fp:
                fp.write(app)

        signed_app = ws.sign_application_request(app)
        if verbose:
            logger.info('------------------------------------------------------ {} signed_app\n{}'.format(call_str, signed_app.decode()))

        if ws.bank_encryption_cert_file:
            enc_app = ws.encrypt_application_request(signed_app)
            if verbose:
                logger.info('------------------------------------------------------ {} enc_app\n{}'.format(call_str, enc_app.decode()))
        else:
            enc_app = signed_app
            if verbose:
                logger.info('------------------------------------------------------ '
                            '{} enc_app\n(no bank_encryption_cert_file, not encrypting)'.format(call_str))

        b64_app = ws.encode_application_request(enc_app)
        if verbose:
            logger.info('------------------------------------------------------ {} b64_app\n{}'.format(call_str, b64_app.decode()))

        soap_body = get_template('jbank/soap_template.xml').render({
            'soap_call': soap_call,
            'payload': b64_app.decode(),
        })
        if verbose:
            logger.info('------------------------------------------------------ {} soap_body\n{}'.format(call_str, soap_body))

        body_bytes = soap_body.encode()
        envelope = etree.fromstring(body_bytes)
        binary_signature = BinarySignature(ws.signing_key_full_path, ws.signing_cert_full_path)
        soap_headers = {
        }
        envelope, soap_headers = binary_signature.apply(envelope, soap_headers)
        signed_body_bytes = etree.tostring(envelope)
        if verbose:
            logger.info('------------------------------------------------------ {} signed_body_bytes\n{}'.format(call_str, signed_body_bytes))

        http_headers = {
            'Connection': 'Close',
            'Content-Type': 'text/xml',
            'Method': 'POST',
            'SOAPAction': '',
            'User-Agent': 'Kajala WS',
        }
        if verbose:
            logger.info('HTTP POST {}'.format(ws.soap_endpoint))
        res = requests.post(ws.soap_endpoint, data=signed_body_bytes, headers=http_headers)
        if verbose:
            logger.info('------------------------------------------------------ {} HTTP response {}\n{}'.format(call_str, res.status_code, res.text))
        if res.status_code >= 300:
            raise Exception("WS-EDI {} HTTP {}".format(command, res.status_code))

        envelope = etree.fromstring(res.content)
        app_res_el = envelope.find('.//{http://model.bxd.fi}ApplicationResponse')
        if app_res_el is None:
            raise Exception("WS-EDI {} failed, missing ApplicationResponse".format(command))
        app_res_enc = ws.decode_application_response(app_res_el.text.encode())
        if verbose:
            logger.info('------------------------------------------------------ {} app_res_enc\n{}'.format(call_str, app_res_enc.decode()))

        if ws.encryption_key_file:
            app_res = ws.decrypt_application_response(app_res_enc)
            if verbose:
                logger.info('------------------------------------------------------ {} app_res\n{}'.format(call_str, app_res.decode()))
        else:
            app_res = app_res_enc
            if verbose:
                logger.info('------------------------------------------------------ '
                            '{} app_res\n(no encryption_key_file, assuming decrypted content)'.format(call_str))

        soap_call.executed = now()
        soap_call.save(update_fields=['executed'])

        if debug_output:
            with open(soap_call.debug_application_response_full_path, 'wb') as fp:
                fp.write(app_res)

        return app_res
    except Exception:
        soap_call.error = traceback.format_exc()
        soap_call.save(update_fields=['error'])
        raise



#pylint: disable=too-many-branches
import logging
import os
from pprint import pprint
from django.core.files import File
from django.core.management import CommandParser
from django.db import transaction
from jbank.helpers import create_statement, get_or_create_bank_account
from jbank.files import list_dir_files
from jbank.models import Statement, StatementFile
from jbank.parsers import parse_tiliote_statements_from_file
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Parses bank account statement .TO (tiliote) files'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('path', type=str)
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--test', action='store_true')
        parser.add_argument('--delete-old', action='store_true')
        parser.add_argument('--auto-create-accounts', action='store_true')
        parser.add_argument('--resolve-original-filenames', action='store_true')
        parser.add_argument('--tag', type=str, default='')

    def do(self, *args, **options):
        files = list_dir_files(options['path'])
        for filename in files:
            plain_filename = os.path.basename(filename)

            if options['resolve_original_filenames']:
                found = StatementFile.objects.filter(statement__name=plain_filename).first()
                if found and not found.original_filename:
                    assert isinstance(found, StatementFile)
                    found.original_filename = filename
                    found.save(update_fields=['original_filename'])
                    logger.info('Original TO statement filename of %s resolved to %s', found, filename)

            if options['delete_old']:
                Statement.objects.filter(name=plain_filename).delete()

            if options['test']:
                statements = parse_tiliote_statements_from_file(filename)
                pprint(statements)
                continue

            if not Statement.objects.filter(name=plain_filename).first():
                print('Importing statement file {}'.format(filename))

                statements = parse_tiliote_statements_from_file(filename)
                if options['verbose']:
                    pprint(statements)

                with transaction.atomic():
                    if not Statement.objects.filter(name=plain_filename).first():
                        file = StatementFile(original_filename=filename, tag=options['tag'])
                        file.save()
                        with open(filename, 'rb') as fp:
                            file.file.save(plain_filename, File(fp))
                        for data in statements:
                            if options['auto_create_accounts']:
                                account_number = data.get('header', {}).get('account_number')
                                if account_number:
                                    get_or_create_bank_account(account_number)

                            create_statement(data, name=plain_filename, file=file)
            else:
                print('Skipping statement file {}'.format(filename))



import os
import subprocess
from datetime import date, datetime

from decimal import Decimal
from os.path import join
from pprint import pprint

import pytz
from django.conf import settings
from django.template.loader import get_template
from django.test import TestCase

from jbank.ecb import parse_euro_exchange_rates_xml
from jbank.helpers import validate_xml
from jbank.models import WsEdiConnection, WsEdiSoapCall
from jbank.parsers import parse_tiliote_statements_from_file, parse_svm_batches_from_file
from jbank.sepa import Pain001, Pain002, PAIN001_REMITTANCE_INFO_OCR, PAIN001_REMITTANCE_INFO_OCR_ISO
from jutil.format import format_xml
from jutil.validators import iban_bic


class Tests(TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_pain001(self):
        debtor_acc = 'FI4947300010416310'
        p = Pain001('201802071211XJANITEST', 'Vuokrahelppi', debtor_acc, iban_bic(debtor_acc), '020840699',
                    ['Koukkukankareentie 29', '20320 Turku'], 'FI')
        creditor_acc = 'FI8847304720017517'
        p.add_payment('201802071339A0001', 'Jani Kajala', creditor_acc, iban_bic(creditor_acc), Decimal('49.00'), 'vuokratilitys')
        p.add_payment('201802071339A0001', 'Jani Kajala', creditor_acc, iban_bic(creditor_acc), Decimal('49.00'), '302300', PAIN001_REMITTANCE_INFO_OCR)
        p.add_payment('201802071339A0001', 'Jani Kajala', creditor_acc, iban_bic(creditor_acc), Decimal('49.00'), 'RF92 1229', PAIN001_REMITTANCE_INFO_OCR_ISO)
        xml_str = format_xml(p.render_to_bytes().decode())
        # print(xml_str)

        filename = '/tmp/pain001.xml'
        with open(filename, 'wt') as fp:
            fp.write(xml_str)
            # print(filename, 'written')

        # /usr/bin/xmllint --format --pretty 1 --load-trace --debug --schema $1 $2
        res = subprocess.run([
            '/usr/bin/xmllint',
            '--noout',
            # '--format',
            # '--pretty', '1',
            # '--load-trace',
            # '--debug',
            '--schema',
            join(settings.BASE_DIR, 'data/pain001/pain.001.001.03.xsd'),
            filename,
        ])
        self.assertEqual(res.returncode, 0)

    def test_to(self):
        filename = join(settings.BASE_DIR, 'data/to/547404896.TO')
        statements = parse_tiliote_statements_from_file(filename)
        rec = statements[0]['records'][0]
        # pprint(rec)
        self.assertEqual(rec['amount'], Decimal('-1799.00'))
        self.assertEqual(rec['archive_identifier'], '180203473047IE5807')
        self.assertEqual(rec['paid_date'], date(2018, 2, 3))
        self.assertEqual(rec['sepa']['iban_account_number'], 'FI8847304720017517')

    def test_svm(self):
        filename = join(settings.BASE_DIR, 'data/svm/547392460.SVM')
        batches = parse_svm_batches_from_file(filename)
        recs = batches[0]['records']
        self.assertEqual(len(recs), 1)
        rec = recs[0]
        # pprint(rec)
        self.assertEqual(rec['amount'], Decimal('49.00'))
        self.assertEqual(rec['archive_identifier'], '02042588WWRV0212')
        self.assertEqual(rec['remittance_info'], '00000000000000013013')

    def test_xp(self):
        filename = join(settings.BASE_DIR, 'data/xp/547958656.XP')
        with open(filename, 'rb') as fp:
            file_content = fp.read()
        p = Pain002(file_content)
        self.assertEqual(p.original_msg_id, '201802071211XJANITEST')
        self.assertEqual(p.msg_id, 'V000000009726773')
        self.assertEqual(p.group_status, 'ACCP')
        self.assertEqual(p.is_accepted, True)

    def test_ecb_rates(self):
        filename = join(settings.BASE_DIR, 'data/ecb-rates-2019-08-15.xml')
        with open(filename, 'rt') as fp:
            content = fp.read()
        rates = parse_euro_exchange_rates_xml(content)
        # for record_date, currency, rate in rates[20:21]:
            # print(record_date, currency, rate)
        record_date, currency, rate = rates[20]
        self.assertEqual(record_date, date(2019, 8, 15))
        self.assertEqual(currency, 'HKD')
        self.assertEqual(rate, Decimal('8.744'))

    def normalize_soap_env(self, content: bytes) -> bytes:
        from lxml import etree  # local
        doc = etree.fromstring(content)
        doc.find('.//{http://www.w3.org/2000/09/xmldsig#}DigestValue').text = 'x'
        doc.find('.//{http://www.w3.org/2000/09/xmldsig#}Reference').attrib['URI'] = 'x'
        doc.find('.//{http://www.w3.org/2000/09/xmldsig#}SignatureValue').text = 'x'
        doc.find('.//{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}Reference').attrib['URI'] = 'x'
        doc.find('.//{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}BinarySecurityToken').attrib['{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd}Id'] = 'x'
        doc.find('.//{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}BinarySecurityToken').text = 'x'
        doc.find('.//{http://schemas.xmlsoap.org/soap/envelope/}Body').attrib['{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd}Id'] = 'x'
        return etree.tostring(doc)

    def test_x509(self):
        from lxml import etree  # local
        from zeep.wsse import BinarySignature
        print('MEDIA_ROOT = {}'.format(settings.MEDIA_ROOT))
        ws = WsEdiConnection(name='test', sender_identifier='12319203', receiver_identifier='123192031', target_identifier='1', environment='TEST', soap_endpoint='http://localhost')
        ws.signing_cert_file.name = 'data/x509/cert.pem'
        ws.signing_key_file.name = 'data/x509/key.pem'
        ws.save()
        self.assertEqual(WsEdiConnection.objects.get_by_target_identifier('1').id, ws.id)
        app = open('data/x509/appreq.xml', 'rb').read()
        signed = ws.sign_application_request(app)
        ref_signed = b'<?xml version="1.0"?>\n<ApplicationRequest xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns="http://bxd.fi/xmldata/">\n    <CustomerId>061133</CustomerId>\n    <Timestamp>2012-02-20T08:50:59.4319012+01:00</Timestamp>\n    <Environment>TEST</Environment>\n    <FileReferences>\n    <FileReference>1202170046-1202171334</FileReference>\n    </FileReferences>\n    <SoftwareId>DBSEPAClient</SoftwareId>\n    <FileType>pain.002.001.02</FileType>\n    <Signature xmlns="http://www.w3.org/2000/09/xmldsig#">\n        <SignedInfo>\n          <CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>\n          <SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"/>\n          <Reference URI="">\n            <Transforms>\n              <Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>\n            </Transforms>\n            <DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>\n            <DigestValue>D4Bg9OJdQHgsXMQM2ecyoqNwCn4=</DigestValue>\n          </Reference>\n        </SignedInfo>\n        <SignatureValue>NwocqJ57CuLnQJaIA0Jm8JM2WlNOrNSM65Qt9nq69az4QMCgZBwRIGudZ+uMVWu0\n8TRriTHgrCTw5W50cVnhC/z+sw0mcNuYskXmGpr+4nUASYbcbT6EzktrHHkbhVTZ\nI4X4NdN8skU06TdgzF1Jvw/GzKENrm4l9hBLGbQyIaJppGmK1lN0g1suxnyvzgZK\ngFEEJZdFgFWcP+mCU+88FkcYOVjoX48rLtr0hSXyWRgUC9x3Seiw5GoS3toaY9dE\nmo6GREWUFEX9TjaH2sOy4WexiejlYisCyEGzBnyBGZ/Xi5EdoZSRoGrh60r+XDLX\n3Vu3vCJm5zD3SPX5fV9o/A==</SignatureValue>\n        <KeyInfo>\n          <X509Data>\n            <X509IssuerSerial>\n              <X509IssuerName>Issuer: {{ ws.signing_cert.issuer.rfc4514_string }}</X509IssuerName>\n              <X509SerialNumber>{{ ws.signing_cert.serial_number }}</X509SerialNumber>\n            </X509IssuerSerial>\n            <X509Certificate>MIIDVDCCAjygAwIBAgIUGhIGnbdNdnLHN2Gb3GCgUeSJjQYwDQYJKoZIhvcNAQEL\nBQAwVzELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAlRYMQ8wDQYDVQQHDAZEYWxsYXMx\nFTATBgNVBAoMDEthamFsYSBHcm91cDETMBEGA1UEAwwKa2FqYWxhLmNvbTAeFw0x\nOTEyMDMxNzU0NDFaFw0xOTEyMTMxNzU0NDFaMFcxCzAJBgNVBAYTAlVTMQswCQYD\nVQQIDAJUWDEPMA0GA1UEBwwGRGFsbGFzMRUwEwYDVQQKDAxLYWphbGEgR3JvdXAx\nEzARBgNVBAMMCmthamFsYS5jb20wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK\nAoIBAQDzpgdgMXiW9SOxnFPLdeFIltlCxH2rAz/nj7bLGK9ycVMMOW08M55jLhMK\nfOnyrHECWSP1aSKV6ZsNrRqm87JK8LEHgnIIDauqmJY4bEqB9nj7SzhAJWw5dJWN\nVNQto0csItaywXm0OjYRo4spDoKlyUbgXEcIouH4nWe91fvpvCfZiTHlfToTqKJy\nHtJ+kPi2+PSySs4+167g3A+v4YnoHgxTJD+q6KSimQQ0VGpqcV4Mt7+U3VV3Jb61\nSJNoanB31DLDugTjCOYm6PfT1/abQgPi7TOkizNa6HIMtMO/KFXC/6P+Bg05cQM9\n1Qq7iZR2zIg62AjC7z41xivdf8PdAgMBAAGjGDAWMBQGA1UdEQQNMAuCCWxvY2Fs\naG9zdDANBgkqhkiG9w0BAQsFAAOCAQEAAn4D/QY8pdvbPYx+yDeYPlHnYv68ErBK\n7Ib2rrtM7jtumBVL9BCneacjdsLmsrXwNdQkQMynxl6bMa+uR3YkXyQSVs+aSwKy\nIkz++rI5ALRI5KQr/DGzWrrmlIbBrXtQkLUR2mnyw9t+ozSMPtdedVCr50c88B5j\ndnksF6odkXet2gpVa5aZ8T2HUl+DtixkKoQ66Ra0/cXXdi3pk6zfRAm8/wtVIzQY\nX2+rur2NOPUuCzdWAvNjzuWCgmH8A2BxCOWBAMJ/GpsVJpqJp0tgm7ah1N+r1y0c\nRqxXZw7bu3NXedB1YqmOvYRcAGK2WXH4aV7I/rDRCc+aMs1GCOocbw==</X509Certificate>\n          </X509Data>\n        </KeyInfo>\n    </Signature>\n</ApplicationRequest>\n'
        self.assertEqual(signed, ref_signed)
        encoded = ws.encode_application_request(signed)
        ref_encoded = b'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'
        self.assertEqual(encoded, ref_encoded)
        timestamp = pytz.timezone('Europe/Helsinki').localize(datetime(2015, 2, 3, 14, 30))
        soap_call = WsEdiSoapCall.objects.create(connection=ws, command='HelloWorld', created=timestamp)
        soap_body = get_template('jbank/soap_template.xml').render({'soap_call': soap_call, 'payload': encoded.decode()})
        body_bytes = soap_body.encode()
        envelope = etree.fromstring(body_bytes)
        binary_signature = BinarySignature(ws.signing_key_full_path, ws.signing_cert_full_path)
        soap_headers = {
        }
        envelope, soap_headers = binary_signature.apply(envelope, soap_headers)
        signed_body_bytes = etree.tostring(envelope)
        ref_bytes = b'<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns1="http://model.bxd.fi" xmlns:ns2="http://bxd.fi/CorporateFileService">\n  <SOAP-ENV:Header><wsse:Security xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"><Signature xmlns="http://www.w3.org/2000/09/xmldsig#">\n<SignedInfo>\n<CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>\n<SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"/>\n<Reference URI="#id-311224dc-3b4c-4729-b057-b3abb3bcd95d">\n<Transforms>\n<Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>\n</Transforms>\n<DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>\n<DigestValue>FGoo2ZirnTBuQ9N22TPlCO7eXaw=</DigestValue>\n</Reference>\n</SignedInfo>\n<SignatureValue>MNseos+F09uzlAH+1B1o01OBPslz80vJZWmi4yJ4NniZBXAD0rtPBRhbVwF4gSw1\nbLt1Eb6YKDl+nyRMb0f1H/QUuIibMXvlKu2IVYyTUktovW29oKkmXKfJtix9Eh+w\nzq4XEDZ3BjbkHW7FrS5ZZGRJB+IeePTHSZZ1JSxiBnOg5BQ0MlypUhdEaaHmJUM+\n7vSp/pttXc8vNdB/8pFCV0Yw/DrjANeNbFv2HXG0p1R1Dhmk2Tj3+cWDpWoFuWoj\nOg1y0eoTHW/5SSob9S9mEQPJi6aYr1ni89dF1I78o7kRDxK/JbyTNBq0YnLaXA2m\n6hJlwwfj9iYH3AkYe9YXsw==</SignatureValue>\n<KeyInfo>\n<wsse:SecurityTokenReference><wsse:Reference ValueType="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3" URI="#id-e328e4ba-0ee6-4570-8476-89c88611e5ac"/></wsse:SecurityTokenReference></KeyInfo>\n</Signature><wsse:BinarySecurityToken xmlns:ns1="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd" ValueType="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3" EncodingType="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary" ns1:Id="id-e328e4ba-0ee6-4570-8476-89c88611e5ac">MIIDVDCCAjygAwIBAgIUGhIGnbdNdnLHN2Gb3GCgUeSJjQYwDQYJKoZIhvcNAQEL\nBQAwVzELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAlRYMQ8wDQYDVQQHDAZEYWxsYXMx\nFTATBgNVBAoMDEthamFsYSBHcm91cDETMBEGA1UEAwwKa2FqYWxhLmNvbTAeFw0x\nOTEyMDMxNzU0NDFaFw0xOTEyMTMxNzU0NDFaMFcxCzAJBgNVBAYTAlVTMQswCQYD\nVQQIDAJUWDEPMA0GA1UEBwwGRGFsbGFzMRUwEwYDVQQKDAxLYWphbGEgR3JvdXAx\nEzARBgNVBAMMCmthamFsYS5jb20wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK\nAoIBAQDzpgdgMXiW9SOxnFPLdeFIltlCxH2rAz/nj7bLGK9ycVMMOW08M55jLhMK\nfOnyrHECWSP1aSKV6ZsNrRqm87JK8LEHgnIIDauqmJY4bEqB9nj7SzhAJWw5dJWN\nVNQto0csItaywXm0OjYRo4spDoKlyUbgXEcIouH4nWe91fvpvCfZiTHlfToTqKJy\nHtJ+kPi2+PSySs4+167g3A+v4YnoHgxTJD+q6KSimQQ0VGpqcV4Mt7+U3VV3Jb61\nSJNoanB31DLDugTjCOYm6PfT1/abQgPi7TOkizNa6HIMtMO/KFXC/6P+Bg05cQM9\n1Qq7iZR2zIg62AjC7z41xivdf8PdAgMBAAGjGDAWMBQGA1UdEQQNMAuCCWxvY2Fs\naG9zdDANBgkqhkiG9w0BAQsFAAOCAQEAAn4D/QY8pdvbPYx+yDeYPlHnYv68ErBK\n7Ib2rrtM7jtumBVL9BCneacjdsLmsrXwNdQkQMynxl6bMa+uR3YkXyQSVs+aSwKy\nIkz++rI5ALRI5KQr/DGzWrrmlIbBrXtQkLUR2mnyw9t+ozSMPtdedVCr50c88B5j\ndnksF6odkXet2gpVa5aZ8T2HUl+DtixkKoQ66Ra0/cXXdi3pk6zfRAm8/wtVIzQY\nX2+rur2NOPUuCzdWAvNjzuWCgmH8A2BxCOWBAMJ/GpsVJpqJp0tgm7ah1N+r1y0c\nRqxXZw7bu3NXedB1YqmOvYRcAGK2WXH4aV7I/rDRCc+aMs1GCOocbw==</wsse:BinarySecurityToken></wsse:Security></SOAP-ENV:Header><SOAP-ENV:Body xmlns:ns0="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd" ns0:Id="id-311224dc-3b4c-4729-b057-b3abb3bcd95d">\n    <ns2:helloWorldin>\n      <ns1:RequestHeader>\n        <ns1:SenderId>12319203</ns1:SenderId>\n        <ns1:RequestId>1</ns1:RequestId>\n        <ns1:Timestamp>2015-02-03T14:30:00+02:00</ns1:Timestamp>\n        <ns1:Language>FI</ns1:Language>\n        <ns1:UserAgent>Kajala WS</ns1:UserAgent>\n        <ns1:ReceiverId>123192031</ns1:ReceiverId>\n      </ns1:RequestHeader>\n    <ns1:ApplicationRequest>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</ns1:ApplicationRequest>\n    </ns2:helloWorldin>\n  </SOAP-ENV:Body>\n</SOAP-ENV:Envelope>'
        self.assertEqual(self.normalize_soap_env(ref_bytes), self.normalize_soap_env(signed_body_bytes))

    def test_validate_xml(self):
        xml = os.path.join(settings.BASE_DIR, 'data/finvoice/xsd-test.xml')
        xsd = os.path.join(settings.BASE_DIR, 'data/finvoice/xsd-test.xsd')
        with open(xml, 'rb') as fp:
            validate_xml(fp.read(), xsd)



# Generated by Django 2.2.3 on 2019-11-28 22:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0043_wsedisoapcall'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='wsedisoapcall',
            name='request_identifier',
        ),
        migrations.RemoveField(
            model_name='wsedisoapcall',
            name='timestamp',
        ),
    ]



import os
import subprocess
from django.core.management import CommandParser
from jutil.command import SafeCommand
import jbank


class Command(SafeCommand):
    help = 'Compiles xmlsec1-examples'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--clean', action='store_true')
        parser.add_argument('--clean-only', action='store_true')

    def do(self, *args, **options):
        package_path = os.path.dirname(jbank.__file__)
        xmlsec1_examples_path = os.path.join(package_path, 'xmlsec1-examples')
        print('xmlsec1-examples @ {}'.format(xmlsec1_examples_path))
        os.chdir(xmlsec1_examples_path)
        if options['clean'] or options['clean_only']:
            subprocess.run(['make', 'clean'], check=True)
        if not options['clean_only']:
            subprocess.run(['make'], check=True)



# Generated by Django 2.2.3 on 2019-11-27 01:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0036_wsediconnection'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='wsediconnection',
            name='customer_id',
        ),
        migrations.AddField(
            model_name='wsediconnection',
            name='customer',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, to='jbank.PayoutParty', verbose_name='customer'),
            preserve_default=False,
        ),
    ]



#pylint: disable=too-many-branches,logging-format-interpolation
import logging
import os
from pprint import pprint
from django.core.files import File
from django.core.management import CommandParser
from django.db import transaction
from jbank.helpers import create_reference_payment_batch, get_or_create_bank_account
from jbank.files import list_dir_files
from jbank.models import ReferencePaymentBatch, ReferencePaymentBatchFile
from jbank.parsers import parse_svm_batches_from_file
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Parses bank settlement .SVM (saapuvat viitemaksut) files'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('path', type=str)
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--test', action='store_true')
        parser.add_argument('--delete-old', action='store_true')
        parser.add_argument('--auto-create-accounts', action='store_true')
        parser.add_argument('--resolve-original-filenames', action='store_true')
        parser.add_argument('--tag', type=str, default='')

    def do(self, *args, **options):
        files = list_dir_files(options['path'])
        # pprint(files)
        for filename in files:
            plain_filename = os.path.basename(filename)

            if options['resolve_original_filenames']:
                found = ReferencePaymentBatchFile.objects.filter(referencepaymentbatch__name=plain_filename).first()
                if found and not found.original_filename:
                    assert isinstance(found, ReferencePaymentBatchFile)
                    found.original_filename = filename
                    found.save(update_fields=['original_filename'])
                    logger.info('Original SVM reference payment batch filename of %s resolved to %s', found, filename)

            if options['delete_old']:
                ReferencePaymentBatch.objects.filter(name=plain_filename).delete()

            if options['test']:
                batches = parse_svm_batches_from_file(filename)
                pprint(batches)
                continue

            if not ReferencePaymentBatch.objects.filter(name=plain_filename).first():
                print('Importing statement file {}'.format(filename))

                batches = parse_svm_batches_from_file(filename)
                if options['verbose']:
                    pprint(batches)

                with transaction.atomic():
                    if not ReferencePaymentBatch.objects.filter(name=plain_filename).first():
                        file = ReferencePaymentBatchFile(original_filename=filename, tag=options['tag'])
                        file.save()
                        with open(filename, 'rb') as fp:
                            file.file.save(plain_filename, File(fp))

                        for data in batches:
                            if options['auto_create_accounts']:
                                for rec_data in data['records']:
                                    account_number = rec_data.get('account_number')
                                    if account_number:
                                        get_or_create_bank_account(account_number)

                            create_reference_payment_batch(data, name=plain_filename, file=file)
            else:
                print('Skipping reference payment file {}'.format(filename))



#pylint: disable=logging-format-interpolation
import logging
from django.core.management import CommandParser
from jbank.models import ReferencePaymentBatchFile, ReferencePaymentRecord
from jbank.parsers import parse_svm_batches_from_file
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Re-parses old bank settlement .SVM (saapuvat viitemaksut) files. Used for adding missing fields.'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--file', type=str)

    def do(self, *args, **options):
        logger.info('Re-parsing SVM files to update fields')
        qs = ReferencePaymentBatchFile.objects.all()
        if options['file']:
            qs = qs.filter(file=options['file'])
        for file in qs.order_by('id'):
            assert isinstance(file, ReferencePaymentBatchFile)
            logger.info('Processing {} BEGIN'.format(file))
            batches = parse_svm_batches_from_file(file.full_path)
            for batch in batches:
                for e in batch['records']:
                    # check missing line_number
                    e2 = ReferencePaymentRecord.objects.filter(batch__file=file, line_number=0, record_type=e['record_type'],
                                                               account_number=e['account_number'],
                                                               paid_date=e['paid_date'], archive_identifier=e['archive_identifier'],
                                                               remittance_info=e['remittance_info'], payer_name=e['payer_name'],
                                                               currency_identifier=e['currency_identifier'], name_source=e['name_source'],
                                                               correction_identifier=e['correction_identifier'],
                                                               delivery_method=e['delivery_method'], receipt_code=e['receipt_code']).first()
                    if e2:
                        e2.line_number = e['line_number']
                        e2.save()
                        logger.info('Updated {} line number to {}'.format(e2, e2.line_number))
            logger.info('Processing {} END'.format(file))



# Generated by Django 2.2.3 on 2019-11-29 20:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0049_wsediconnection_sender_identifier'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='wsedisoapcall',
            name='payout',
        ),
    ]



# Generated by Django 2.2.3 on 2019-11-28 22:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0044_auto_20191128_2231'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='receiver_identifier',
            field=models.CharField(default='', max_length=32, verbose_name='receiver identifier'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.3 on 2019-08-15 20:59

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0033_auto_20190801_2121'),
    ]

    operations = [
        migrations.CreateModel(
            name='CurrencyExchangeSource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, verbose_name='name')),
                ('created', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created')),
            ],
            options={
                'verbose_name': 'currency exchange source',
                'verbose_name_plural': 'currency exchange sources',
            },
        ),
        migrations.AlterModelOptions(
            name='currencyexchange',
            options={'verbose_name': 'currency exchange', 'verbose_name_plural': 'currency exchanges'},
        ),
        migrations.AddField(
            model_name='currencyexchange',
            name='source',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.PROTECT, to='jbank.CurrencyExchangeSource', verbose_name='currency exchange source'),
        ),
    ]



# Generated by Django 2.2.3 on 2019-11-28 23:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0047_wsedisoapcall_error'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='soap_endpoint',
            field=models.URLField(default='https://businessws.danskebank.com/financialservice/edifileservice.asmx', verbose_name='SOAP endpoint'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.3 on 2019-12-03 00:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0055_auto_20191130_2011'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='debug_commands',
            field=models.TextField(blank=True, verbose_name='debug commands'),
        ),
    ]



# Generated by Django 2.2.8 on 2020-03-16 19:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0057_wsediconnection_enabled'),
    ]

    operations = [
        migrations.AlterField(
            model_name='statementrecordremittanceinfo',
            name='additional_info',
            field=models.CharField(blank=True, db_index=True, max_length=256, verbose_name='additional remittance info'),
        ),
    ]



import base64
import logging
import os
import re
import subprocess
import tempfile
from datetime import datetime, time
from os.path import basename, join
from pathlib import Path
from typing import List
import cryptography
import pytz
from cryptography import x509
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.template.loader import get_template
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from jacc.models import AccountEntry, AccountEntrySourceFile, Account
from jutil.dict import choices_label
from jutil.format import format_xml
from jutil.validators import iban_validator, iban_bic, iso_payment_reference_validator, fi_payment_reference_validator


logger = logging.getLogger(__name__)


JBANK_BIN_PATH = Path(__file__).absolute().parent.joinpath('bin')

RECORD_ENTRY_TYPE = (
    ('1', _('Deposit')),
    ('2', _('Withdrawal')),
    ('3', _('Deposit Correction')),
    ('4', _('Withdrawal Correction')),
)

RECORD_CODES = (
    ('700', _('Money Transfer (In/Out)')),
    ('701', _('Recurring Payment (In/Out)')),
    ('702', _('Bill Payment (Out)')),
    ('703', _('Payment Terminal Deposit (In)')),
    ('704', _('Bank Draft (In/Out)')),
    ('705', _('Reference Payments (In)')),
    ('706', _('Payment Service (Out)')),
    ('710', _('Deposit (In)')),
    ('720', _('Withdrawal (Out)')),
    ('721', _('Card Payment (Out)')),
    ('722', _('Check (Out)')),
    ('730', _('Bank Fees (Out)')),
    ('740', _('Interests Charged (Out)')),
    ('750', _('Interests Credited (In)')),
    ('760', _('Loan (Out)')),
    ('761', _('Loan Payment (Out)')),
    ('770', _('Foreign Transfer (In/Out)')),
    ('780', _('Zero Balancing (In/Out)')),
    ('781', _('Sweeping (In/Out)')),
    ('782', _('Topping (In/Out)')),
)

RECORD_DOMAIN = (
    ('PMNT', _('Money Transfer (In/Out)')),
    ('LDAS', _('Loan Payment (Out)')),
    ('CAMT', _('Cash Management')),
    ('ACMT', _('Account Management')),
    ('XTND', _('Entended Domain')),
    ('SECU', _('Securities')),
    ('FORX', _('Foreign Exchange')),
    ('XTND', _('Entended Domain')),
    ('NTAV', _('Not Available'))
)

RECEIPT_CODE = (
    ('', ''),
    ('E', _('Separate')),
    ('P', _('Separate/Paper')),
)

CURRENCY_IDENTIFIERS = (
    ('1', 'EUR'),
)

NAME_SOURCES = (
    ('', _('Not Set')),
    ('A', _('From Customer')),
    ('K', _('From Bank Clerk')),
    ('J', _('From Bank System')),
)

CORRECTION_IDENTIFIER = (
    ('0', _('Regular Entry')),
    ('1', _('Correction Entry')),
)

DELIVERY_FROM_CUSTOMER = 'A'
DELIVERY_FROM_BANK_CLERK = 'K'
DELIVERY_FROM_BANK_SYSTEM = 'J'

DELIVERY_METHOD = (
    (DELIVERY_FROM_CUSTOMER, _('From Customer')),
    (DELIVERY_FROM_BANK_CLERK, _('From Bank Clerk')),
    (DELIVERY_FROM_BANK_SYSTEM, _('From Bank System')),
)

PAYOUT_WAITING_PROCESSING = 'W'
PAYOUT_WAITING_UPLOAD = 'U'
PAYOUT_UPLOADED = 'D'
PAYOUT_PAID = 'P'
PAYOUT_CANCELED = 'C'
PAYOUT_ERROR = 'E'

PAYOUT_STATE = (
    (PAYOUT_WAITING_PROCESSING, _('waiting processing')),
    (PAYOUT_WAITING_UPLOAD, _('waiting upload')),
    (PAYOUT_UPLOADED, _('uploaded')),
    (PAYOUT_PAID, _('paid')),
    (PAYOUT_CANCELED, _('canceled')),
    (PAYOUT_ERROR, _('error')),
)


class Statement(AccountEntrySourceFile):
    file = models.ForeignKey('StatementFile', blank=True, default=None, null=True, on_delete=models.CASCADE)
    account = models.ForeignKey(Account, related_name='+', on_delete=models.PROTECT)
    account_number = models.CharField(_('account number'), max_length=32, db_index=True)
    statement_identifier = models.CharField(_('statement identifier'), max_length=48, db_index=True, blank=True, default='')
    statement_number = models.SmallIntegerField(_('statement number'), db_index=True)
    begin_date = models.DateField(_('begin date'), db_index=True)
    end_date = models.DateField(_('end date'), db_index=True)
    record_date = models.DateTimeField(_('record date'), db_index=True)
    customer_identifier = models.CharField(_('customer identifier'), max_length=64, blank=True, default='')
    begin_balance_date = models.DateField(_('begin balance date'), )
    begin_balance = models.DecimalField(_('begin balance'), max_digits=10, decimal_places=2)
    record_count = models.IntegerField(_('record count'), null=True, default=None)
    currency_code = models.CharField(_('currency code'), max_length=3)
    account_name = models.CharField(_('account name'), max_length=32, blank=True, default='')
    account_limit = models.DecimalField(_('account limit'), max_digits=10, decimal_places=2, blank=True, default=None, null=True)
    owner_name = models.CharField(_('owner name'), max_length=64)
    contact_info_1 = models.CharField(_('contact info (1)'), max_length=64, blank=True, default='')
    contact_info_2 = models.CharField(_('contact info (2)'), max_length=64, blank=True, default='')
    bank_specific_info_1 = models.CharField(_('bank specific info (1)'), max_length=1024, blank=True, default='')
    iban = models.CharField(_('IBAN'), max_length=32, db_index=True)
    bic = models.CharField(_('BIC'), max_length=11, db_index=True)

    class Meta:
        verbose_name = _('statement')
        verbose_name_plural = _('statements')


class PaymentRecordManager(models.Manager):
    def filter_matched(self):
        return self.exclude(child_set=None)

    def filter_unmatched(self):
        return self.filter(child_set=None)


class StatementRecord(AccountEntry):
    objects = PaymentRecordManager()
    statement = models.ForeignKey(Statement, verbose_name=_('statement'), related_name='record_set', on_delete=models.CASCADE)
    line_number = models.SmallIntegerField(_('line number'), default=None, null=True, blank=True)
    record_number = models.IntegerField(_('record number'), default=None, null=True, blank=True)
    archive_identifier = models.CharField(_('archive identifier'), max_length=64, blank=True, default='', db_index=True)
    record_date = models.DateField(_('record date'), db_index=True)
    value_date = models.DateField(_('value date'), db_index=True, blank=True, null=True, default=None)
    paid_date = models.DateField(_('paid date'), db_index=True, blank=True, null=True, default=None)
    entry_type = models.CharField(_('entry type'), max_length=1, choices=RECORD_ENTRY_TYPE, db_index=True)
    record_code = models.CharField(_('record type'), max_length=4, choices=RECORD_CODES, db_index=True, blank=True)
    record_domain = models.CharField(_('record domain'), max_length=4, choices=RECORD_DOMAIN, db_index=True, blank=True)
    family_code = models.CharField(_('family code'), max_length=4, db_index=True, blank=True, default='')
    sub_family_code = models.CharField(_('sub family code'), max_length=4, db_index=True, blank=True, default='')
    record_description = models.CharField(_('record description'), max_length=128, blank=True, default='')
    receipt_code = models.CharField(_('receipt code'), max_length=1, choices=RECEIPT_CODE, db_index=True, blank=True)
    delivery_method = models.CharField(_('delivery method'), max_length=1, db_index=True, choices=DELIVERY_METHOD)
    name = models.CharField(_('name'), max_length=64, blank=True, db_index=True)
    name_source = models.CharField(_('name source'), max_length=1, blank=True, choices=NAME_SOURCES)
    recipient_account_number = models.CharField(_('recipient account number'), max_length=32, blank=True, db_index=True)
    recipient_account_number_changed = models.CharField(_('recipient account number changed'), max_length=1, blank=True)
    remittance_info = models.CharField(_('remittance info'), max_length=35, db_index=True, blank=True)
    messages = models.TextField(_('messages'), blank=True, default='')
    client_messages = models.TextField(_('client messages'), blank=True, default='')
    bank_messages = models.TextField(_('bank messages'), blank=True, default='')
    manually_settled = models.BooleanField(_('manually settled'), db_index=True, default=False, blank=True)

    class Meta:
        verbose_name = _('statement record')
        verbose_name_plural = _('statement records')

    def clean(self):
        self.source_file = self.statement
        self.timestamp = pytz.utc.localize(datetime.combine(self.record_date, time(0, 0)))
        if self.name:
            self.description = '{name}: {record_description}'.format(record_description=self.record_description, name=self.name)
        else:
            self.description = '{record_description}'.format(record_description=self.record_description)


class CurrencyExchangeSource(models.Model):
    name = models.CharField(_('name'), max_length=64)
    created = models.DateTimeField(_('created'), default=now, db_index=True, blank=True, editable=False)

    class Meta:
        verbose_name = _('currency exchange source')
        verbose_name_plural = _('currency exchange sources')

    def __str__(self):
        return self.name


class CurrencyExchange(models.Model):
    record_date = models.DateField(_('record date'), db_index=True)
    source_currency = models.CharField(_('source currency'), max_length=3, blank=True)
    target_currency = models.CharField(_('target currency'), max_length=3, blank=True)
    unit_currency = models.CharField(_('unit currency'), max_length=3, blank=True)
    exchange_rate = models.DecimalField(_('exchange rate'), decimal_places=6, max_digits=12, null=True, default=None, blank=True)
    source = models.ForeignKey(CurrencyExchangeSource, verbose_name=_('currency exchange source'), blank=True, null=True, default=None, on_delete=models.PROTECT)  # noqa

    class Meta:
        verbose_name = _('currency exchange')
        verbose_name_plural = _('currency exchanges')

    def __str__(self):
        return '{src} = {rate} {tgt}'.format(src=self.source_currency, tgt=self.target_currency, rate=self.exchange_rate)


class StatementRecordDetail(models.Model):
    record = models.ForeignKey(StatementRecord, verbose_name=_('record'), related_name='detail_set', on_delete=models.CASCADE)
    batch_identifier = models.CharField(_('batch message id'), max_length=64, db_index=True, blank=True, default='')
    amount = models.DecimalField(verbose_name=_('amount'), max_digits=10, decimal_places=2, blank=True, default=None, null=True, db_index=True)
    currency_code = models.CharField(_('currency code'), max_length=3)
    instructed_amount = models.DecimalField(verbose_name=_('instructed amount'), max_digits=10, decimal_places=2, blank=True, default=None, null=True, db_index=True)    # noqa
    exchange = models.ForeignKey(CurrencyExchange, verbose_name=_('currency exchange'), related_name='recorddetail_set', on_delete=models.PROTECT, null=True, default=None, blank=True)  # noqa
    archive_identifier = models.CharField(_('archive identifier'), max_length=64, blank=True)
    end_to_end_identifier = models.CharField(_('end-to-end identifier'), max_length=64, blank=True)
    creditor_name = models.CharField(_('creditor name'), max_length=128, blank=True)
    creditor_account = models.CharField(_('creditor account'), max_length=35, blank=True)
    debtor_name = models.CharField(_('debtor name'), max_length=128, blank=True)
    ultimate_debtor_name = models.CharField(_('ultimate debtor name'), max_length=128, blank=True)
    unstructured_remittance_info = models.CharField(_('unstructured remittance info'), max_length=2048, blank=True)
    paid_date = models.DateTimeField(_('paid date'), db_index=True, blank=True, null=True, default=None)


class StatementRecordRemittanceInfo(models.Model):
    detail = models.ForeignKey(StatementRecordDetail, related_name='remittanceinfo_set', on_delete=models.CASCADE)
    additional_info = models.CharField(_('additional remittance info'), max_length=256, blank=True, db_index=True)
    amount = models.DecimalField(_('amount'), decimal_places=2, max_digits=10, null=True, default=None, blank=True)
    currency_code = models.CharField(_('currency code'), max_length=3, blank=True)
    reference = models.CharField(_('reference'), max_length=35, blank=True, db_index=True)

    def __str__(self):
        return '{} {} ref {} ({})'.format(self.amount if self.amount is not None else '', self.currency_code, self.reference, self.additional_info)


class StatementRecordSepaInfo(models.Model):
    record = models.OneToOneField(StatementRecord, verbose_name=_('record'), related_name='sepa_info', on_delete=models.CASCADE)
    reference = models.CharField(_('reference'), max_length=35, blank=True)
    iban_account_number = models.CharField(_('IBAN'), max_length=35, blank=True)
    bic_code = models.CharField(_('BIC'), max_length=35, blank=True)
    recipient_name_detail = models.CharField(_('recipient name detail'), max_length=70, blank=True)
    payer_name_detail = models.CharField(_('payer name detail'), max_length=70, blank=True)
    identifier = models.CharField(_('identifier'), max_length=35, blank=True)
    archive_identifier = models.CharField(_('archive identifier'), max_length=64, blank=True)

    class Meta:
        verbose_name = _('SEPA')
        verbose_name_plural = _('SEPA')

    def __str__(self):
        return '[{}]'.format(self.id)


class ReferencePaymentBatchManager(models.Manager):
    def latest_record_date(self) -> datetime:
        """
        :return: datetime of latest record available or None
        """
        obj = self.order_by('-record_date').first()
        if not obj:
            return None
        return obj.record_date


class ReferencePaymentBatch(AccountEntrySourceFile):
    objects = ReferencePaymentBatchManager()
    file = models.ForeignKey('ReferencePaymentBatchFile', blank=True, default=None, null=True, on_delete=models.CASCADE)
    record_date = models.DateTimeField(_('record date'), db_index=True)
    institution_identifier = models.CharField(_('institution identifier'), max_length=2, blank=True)
    service_identifier = models.CharField(_('service identifier'), max_length=9, blank=True)
    currency_identifier = models.CharField(_('currency identifier'), max_length=3, choices=CURRENCY_IDENTIFIERS)

    class Meta:
        verbose_name = _('reference payment batch')
        verbose_name_plural = _('reference payment batches')


class ReferencePaymentRecord(AccountEntry):
    """
    Reference payment record. See jacc.Invoice for date/time variable naming conventions.
    """
    objects = PaymentRecordManager()
    batch = models.ForeignKey(ReferencePaymentBatch, verbose_name=_('batch'), related_name='record_set', on_delete=models.CASCADE)
    line_number = models.SmallIntegerField(_('line number'), default=0, blank=True)
    record_type = models.CharField(_('record type'), max_length=1)
    account_number = models.CharField(_('account number'), max_length=32, db_index=True)
    record_date = models.DateField(_('record date'), db_index=True)
    paid_date = models.DateField(_('paid date'), db_index=True, blank=True, null=True, default=None)
    archive_identifier = models.CharField(_('archive identifier'), max_length=32, blank=True, default='', db_index=True)
    remittance_info = models.CharField(_('remittance info'), max_length=32, db_index=True)
    payer_name = models.CharField(_('payer name'), max_length=12, db_index=True)
    currency_identifier = models.CharField(_('currency identifier'), max_length=1, choices=CURRENCY_IDENTIFIERS)
    name_source = models.CharField(_('name source'), max_length=1, choices=NAME_SOURCES, blank=True)
    correction_identifier = models.CharField(_('correction identifier'), max_length=1, choices=CORRECTION_IDENTIFIER)
    delivery_method = models.CharField(_('delivery method'), max_length=1, db_index=True, choices=DELIVERY_METHOD)
    receipt_code = models.CharField(_('receipt code'), max_length=1, choices=RECEIPT_CODE, db_index=True, blank=True)
    manually_settled = models.BooleanField(_('manually settled'), db_index=True, default=False, blank=True)

    class Meta:
        verbose_name = _('reference payment records')
        verbose_name_plural = _('reference payment records')

    @property
    def remittance_info_short(self) -> str:
        """
        Remittance info without preceding zeroes.
        :return: str
        """
        return re.sub(r'^0+', '', self.remittance_info)

    def clean(self):
        self.source_file = self.batch
        self.timestamp = pytz.utc.localize(datetime.combine(self.paid_date, time(0, 0)))
        self.description = '{amount} {remittance_info} {payer_name}'.format(amount=self.amount, remittance_info=self.remittance_info,
                                                                            payer_name=self.payer_name)


class StatementFile(models.Model):
    created = models.DateTimeField(_('created'), default=now, db_index=True, blank=True, editable=False)
    file = models.FileField(_('file'), upload_to='uploads')
    original_filename = models.CharField(_('original filename'), blank=True, default='', max_length=256)
    tag = models.CharField(_('tag'), blank=True, max_length=64, default='', db_index=True)
    errors = models.TextField(_('errors'), max_length=4086, default='', blank=True)

    class Meta:
        verbose_name = _('account statement file')
        verbose_name_plural = _('account statement files')

    @property
    def full_path(self):
        return join(settings.MEDIA_ROOT, self.file.name) if self.file else ''

    def __str__(self):
        return basename(str(self.file.name)) if self.file else ''


class ReferencePaymentBatchFile(models.Model):
    created = models.DateTimeField(_('created'), default=now, db_index=True, blank=True, editable=False)
    file = models.FileField(_('file'), upload_to='uploads')
    original_filename = models.CharField(_('original filename'), blank=True, default='', max_length=256)
    tag = models.CharField(_('tag'), blank=True, max_length=64, default='', db_index=True)
    errors = models.TextField(_('errors'), max_length=4086, default='', blank=True)

    class Meta:
        verbose_name = _("reference payment batch file")
        verbose_name_plural = _("reference payment batch files")

    @property
    def full_path(self):
        return join(settings.MEDIA_ROOT, self.file.name) if self.file else ''

    def __str__(self):
        return basename(str(self.file.name)) if self.file else ''


class PayoutParty(models.Model):
    name = models.CharField(_('name'), max_length=128, db_index=True)
    account_number = models.CharField(_('account number'), max_length=35, db_index=True, validators=[iban_validator])
    bic = models.CharField(_('BIC'), max_length=16, db_index=True, blank=True)
    org_id = models.CharField(_('organization id'), max_length=32, db_index=True, blank=True, default='')
    address = models.TextField(_('address'), blank=True, default='')
    country_code = models.CharField(_('country code'), max_length=2, default='FI', blank=True, db_index=True)
    payouts_account = models.ForeignKey(Account, verbose_name=_('payouts account'), null=True, default=None, blank=True, on_delete=models.PROTECT)

    class Meta:
        verbose_name = _("payout party")
        verbose_name_plural = _("payout parties")

    def __str__(self):
        return '{} ({})'.format(self.name, self.account_number)

    def clean(self):
        if not self.bic:
            self.bic = iban_bic(self.account_number)

    @property
    def address_lines(self):
        out = []
        for line in self.address.split('\n'):
            line = line.strip()
            if line:
                out.append(line)
        return out


class Payout(AccountEntry):
    connection = models.ForeignKey('WsEdiConnection', verbose_name=_('WS-EDI connection'), on_delete=models.SET_NULL, null=True, blank=True, related_name='+')
    payer = models.ForeignKey(PayoutParty, verbose_name=_('payer'), related_name='+', on_delete=models.PROTECT)
    recipient = models.ForeignKey(PayoutParty, verbose_name=_('recipient'), related_name='+', on_delete=models.PROTECT)
    messages = models.TextField(_('recipient messages'), blank=True, default='')
    reference = models.CharField(_('recipient reference'), blank=True, default='', max_length=32)
    msg_id = models.CharField(_('message id'), max_length=64, blank=True, db_index=True, editable=False)
    file_name = models.CharField(_('file name'), max_length=255, blank=True, db_index=True, editable=False)
    full_path = models.TextField(_('full path'), blank=True, editable=False)
    file_reference = models.CharField(_('file reference'), max_length=255, blank=True, db_index=True, editable=False)
    due_date = models.DateField(_('due date'), db_index=True, blank=True, null=True, default=None)
    paid_date = models.DateTimeField(_('paid date'), db_index=True, blank=True, null=True, default=None)
    state = models.CharField(_('state'), max_length=1, blank=True, default=PAYOUT_WAITING_PROCESSING, choices=PAYOUT_STATE, db_index=True)

    class Meta:
        verbose_name = _("payout")
        verbose_name_plural = _("payouts")

    def clean(self):
        if self.parent and not self.amount:
            self.amount = self.parent.amount

        # prevent defining both reference and messages
        if self.messages and self.reference or not self.messages and not self.reference:
            raise ValidationError(_('payment.must.have.reference.or.messages'))

        # validate reference if any
        if self.reference:
            if self.reference[:2] == 'RF':
                iso_payment_reference_validator(self.reference)
            else:
                fi_payment_reference_validator(self.reference)

        # prevent canceling payouts which have been uploaded successfully
        if self.state == PAYOUT_CANCELED:
            if self.is_upload_done:
                group_status = self.group_status
                if group_status != 'RJCT':
                    raise ValidationError(_('File already uploaded') + ' ({})'.format(group_status))

        # save paid time if marking payout as paid manually
        if self.state == PAYOUT_PAID and not self.paid_date:
            self.paid_date = now()
            status = self.payoutstatus_set.order_by('-created').first()
            if status:
                assert isinstance(status, PayoutStatus)
                self.paid_date = status.created

    def generate_msg_id(self, commit: bool = True):
        msg_id_base = re.sub(r'[^\d]', '', now().isoformat())[:-4]
        self.msg_id = msg_id_base + 'P' + str(self.id)
        if commit:
            self.save(update_fields=['msg_id'])

    @property
    def state_name(self):
        return choices_label(PAYOUT_STATE, self.state)

    @property
    def is_upload_done(self):
        return PayoutStatus.objects.filter(payout=self, response_code='00').first() is not None

    @property
    def is_accepted(self):
        return self.has_group_status('ACCP')

    @property
    def is_rejected(self):
        return self.has_group_status('RJCT')

    def has_group_status(self, group_status: str) -> bool:
        return PayoutStatus.objects.filter(payout=self, group_status=group_status).first() is not None

    @property
    def group_status(self):
        status = PayoutStatus.objects.filter(payout=self).order_by('-id').first()
        return status.group_status if status else ''
    group_status.fget.short_description = _('payment.group.status')


class PayoutStatusManager(models.Manager):
    def is_file_processed(self, filename: str) -> bool:
        return self.filter(file_name=basename(filename)).first() is not None


class PayoutStatus(models.Model):
    objects = PayoutStatusManager()
    payout = models.ForeignKey(Payout, verbose_name=_('payout'), related_name='payoutstatus_set', on_delete=models.PROTECT, null=True, default=None, blank=True)  # noqa
    created = models.DateTimeField(_('created'), default=now, db_index=True, editable=False, blank=True)
    file_name = models.CharField(_('file name'), max_length=255, blank=True, db_index=True, editable=False)
    response_code = models.CharField(_('response code'), max_length=4, blank=True, db_index=True)
    response_text = models.CharField(_('response text'), max_length=128, blank=True)
    msg_id = models.CharField(_('message id'), max_length=64, blank=True, db_index=True)
    original_msg_id = models.CharField(_('original message id'), blank=True, max_length=64, db_index=True)
    group_status = models.CharField(_('group status'), max_length=8, blank=True, db_index=True)
    status_reason = models.CharField(_('status reason'), max_length=255, blank=True)

    class Meta:
        verbose_name = _("payout status")
        verbose_name_plural = _("payout statuses")

    def __str__(self):
        return self.group_status

    @property
    def is_accepted(self):
        return self.group_status == 'ACCP'

    @property
    def is_rejected(self):
        return self.group_status == 'RJCT'


class Refund(Payout):
    class Meta:
        verbose_name = _("refund")
        verbose_name_plural = _("refunds")


class WsEdiSoapCall(models.Model):
    connection = models.ForeignKey('WsEdiConnection', verbose_name=_('WS-EDI connection'), on_delete=models.CASCADE)
    command = models.CharField(_('command'), max_length=64, blank=True, db_index=True)
    created = models.DateTimeField(_('created'), default=now, db_index=True, editable=False, blank=True)
    executed = models.DateTimeField(_('executed'), default=None, null=True, db_index=True, editable=False, blank=True)
    error = models.TextField(_('error'), blank=True)

    class Meta:
        verbose_name = _("WS-EDI SOAP call")
        verbose_name_plural = _("WS-EDI SOAP calls")

    def __str__(self):
        return 'WsEdiSoapCall({})'.format(self.id)

    @property
    def timestamp(self) -> datetime:
        return self.created.astimezone(pytz.timezone('Europe/Helsinki'))

    @property
    def request_identifier(self) -> str:
        return str(self.id)

    @property
    def command_camelcase(self) -> str:
        return self.command[0:1].lower() + self.command[1:]

    def debug_get_filename(self, file_type: str) -> str:
        return '{:08}{}.xml'.format(self.id, file_type)

    @property
    def debug_application_request_full_path(self) -> str:
        return self.debug_get_file_path(self.debug_get_filename('a'))

    @property
    def debug_application_response_full_path(self) -> str:
        return self.debug_get_file_path(self.debug_get_filename('r'))

    @staticmethod
    def debug_get_file_path(filename: str) -> str:
        return os.path.join(settings.WSEDI_LOG_PATH, filename) if hasattr(settings, 'WSEDI_LOG_PATH') and settings.WSEDI_LOG_PATH else ''


class WsEdiConnectionManager(models.Manager):
    def get_by_target_identifier(self, target_identifier: str):
        objs = list(self.filter(target_identifier=target_identifier))
        if len(objs) != 1:
            raise ValidationError(_(
                'WS-EDI connection cannot be found by target identifier {target_identifier} since there are {matches} matches').format(
                target_identifier=target_identifier, matches=len(objs)))
        return objs[0]


class WsEdiConnection(models.Model):
    objects = WsEdiConnectionManager()
    name = models.CharField(_('name'), max_length=64)
    enabled = models.BooleanField(_('enabled'), blank=True, default=True)
    sender_identifier = models.CharField(_('sender identifier'), max_length=32)
    receiver_identifier = models.CharField(_('receiver identifier'), max_length=32)
    target_identifier = models.CharField(_('target identifier'), max_length=32)
    environment = models.CharField(_('environment'), max_length=32, default='PRODUCTION')
    soap_endpoint = models.URLField(_('SOAP endpoint'))
    signing_cert_file = models.FileField(_('signing certificate file'), blank=True, upload_to='certs')
    signing_key_file = models.FileField(_('signing key file'), blank=True, upload_to='certs')
    encryption_cert_file = models.FileField(_('encryption certificate file'), blank=True, upload_to='certs')
    encryption_key_file = models.FileField(_('encryption key file'), blank=True, upload_to='certs')
    bank_encryption_cert_file = models.FileField(_('bank encryption cert file'), blank=True, upload_to='certs')
    debug_commands = models.TextField(_('debug commands'), blank=True)
    created = models.DateTimeField(_('created'), default=now, db_index=True, editable=False, blank=True)
    _signing_cert = None

    class Meta:
        verbose_name = _("WS-EDI connection")
        verbose_name_plural = _("WS-EDI connections")

    def __str__(self):
        return '{} / {}'.format(self.name, self.receiver_identifier)

    @property
    def signing_cert_full_path(self) -> str:
        return self.signing_cert_file.file.name if self.signing_cert_file else ''

    @property
    def signing_key_full_path(self) -> str:
        return self.signing_key_file.file.name if self.signing_key_file else ''

    @property
    def encryption_cert_full_path(self) -> str:
        return self.encryption_cert_file.file.name if self.signing_cert_file else ''

    @property
    def encryption_key_full_path(self) -> str:
        return self.encryption_key_file.file.name if self.signing_key_file else ''

    @property
    def bank_encryption_cert_full_path(self) -> str:
        return self.bank_encryption_cert_file.file.name if self.bank_encryption_cert_file else ''

    @property
    def bank_encryption_cert_with_public_key_full_path(self) -> str:
        src_file = self.bank_encryption_cert_full_path
        file = src_file[:-4] + '-with-pubkey.pem'
        if not os.path.isfile(file):
            out = subprocess.check_output([
                settings.OPENSSL_PATH,
                'x509',
                '-pubkey',
                '-in',
                src_file,
            ])
            with open(file, 'wb') as fp:
                fp.write(out)
        return file

    @property
    def signing_cert(self):
        if hasattr(self, '_signing_cert') and self._signing_cert:
            return self._signing_cert
        pem_data = open(self.signing_cert_full_path, 'rb').read()
        cert = x509.load_pem_x509_certificate(pem_data, cryptography.hazmat.backends.default_backend())
        self._signing_cert = cert
        return cert

    def get_application_request(self, command: str, **kwargs) -> bytes:
        return format_xml(get_template('jbank/application_request_template.xml').render({
            'ws': self,
            'command': command,
            'timestamp': now().astimezone(pytz.timezone('Europe/Helsinki')).isoformat(),
            **kwargs
        })).encode()

    def verify_signature(self, content: bytes):
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(content)
            fp.flush()
            subprocess.check_output([
                settings.XMLSEC1_PATH,
                '--verify',
                '--pubkey-pem',
                self.signing_key_full_path,
                fp.name
            ])

    def sign_application_request(self, content: bytes) -> bytes:
        """
        Sign application request.
        See https://users.dcc.uchile.cl/~pcamacho/tutorial/web/xmlsec/xmlsec.html
        :param content: XML application request
        :return: str
        """
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(content)
            fp.flush()
            out = subprocess.check_output([
                settings.XMLSEC1_PATH,
                '--sign',
                '--privkey-pem',
                '{},{}'.format(self.signing_key_full_path, self.signing_cert_full_path),
                fp.name
            ])
        self.verify_signature(out)
        return out

    def encrypt_application_request(self, content: bytes) -> bytes:
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(content)
            fp.flush()
            out = subprocess.check_output([
                self._xmlsec1_example_bin('encrypt3'),
                fp.name,
                self.bank_encryption_cert_with_public_key_full_path,
                self.bank_encryption_cert_full_path
            ])
        return out

    def encode_application_request(self, content: bytes) -> bytes:
        lines = content.split(b'\n')
        if lines and lines[0].startswith(b'<?xml'):
            lines = lines[1:]
        content_without_xml_tag = b'\n'.join(lines)
        return base64.b64encode(content_without_xml_tag)

    def decode_application_response(self, content: bytes) -> bytes:
        return base64.b64decode(content)

    def decrypt_application_response(self, content: bytes) -> bytes:
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(content)
            fp.flush()
            out = subprocess.check_output([
                self._xmlsec1_example_bin('decrypt3'),
                fp.name,
                self.encryption_key_full_path,
            ])
        return out

    @property
    def debug_command_list(self) -> List[str]:
        return [x for x in re.sub(r'[^\w]+', ' ', self.debug_commands).strip().split(' ') if x]

    @staticmethod
    def _xmlsec1_example_bin(file: str) -> str:
        xmlsec1_examples_path = settings.XMLSEC1_EXAMPLES_PATH if hasattr(settings, 'XMLSEC1_EXAMPLES_PATH') and settings.XMLSEC1_EXAMPLES_PATH else ''
        if not xmlsec1_examples_path:
            xmlsec1_examples_path = os.path.join(os.getenv('HOME'), 'bin/xmlsec1-examples')
        return str(os.path.join(xmlsec1_examples_path, file))



# Generated by Django 2.2.8 on 2019-12-20 06:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0056_wsediconnection_debug_commands'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='enabled',
            field=models.BooleanField(blank=True, default=True, verbose_name='enabled'),
        ),
    ]



# Generated by Django 2.2.3 on 2019-11-28 22:25

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0042_wsediconnection_bank_encryption_cert_file'),
    ]

    operations = [
        migrations.CreateModel(
            name='WsEdiSoapCall',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, verbose_name='timestamp')),
                ('request_identifier', models.CharField(blank=True, db_index=True, max_length=64, unique=True, verbose_name='request identifier')),
                ('command', models.CharField(blank=True, db_index=True, max_length=64, verbose_name='command')),
                ('created', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('executed', models.DateTimeField(blank=True, db_index=True, default=None, editable=False, null=True, verbose_name='executed')),
                ('connection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='jbank.WsEdiConnection', verbose_name='WS-EDI connection')),
                ('payout', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='jbank.Payout', verbose_name='payout')),
            ],
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-25 19:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0059_auto_20200324_0234'),
    ]

    operations = [
        migrations.AddField(
            model_name='referencepaymentbatchfile',
            name='original_filename',
            field=models.CharField(blank=True, default='', max_length=256, verbose_name='original filename'),
        ),
        migrations.AddField(
            model_name='statementfile',
            name='original_filename',
            field=models.CharField(blank=True, default='', max_length=256, verbose_name='original filename'),
        ),
    ]



# Generated by Django 2.2.3 on 2019-11-27 01:30

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0035_auto_20190815_2137'),
    ]

    operations = [
        migrations.CreateModel(
            name='WsEdiConnection',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('customer_id', models.CharField(max_length=64, verbose_name='customer id')),
                ('signing_cert', models.FileField(blank=True, upload_to='', verbose_name='signing cert')),
                ('created', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created')),
            ],
            options={
                'verbose_name': 'WS-EDI connection',
                'verbose_name_plural': 'WS-EDI connections',
            },
        ),
    ]



import json
import os
import zipfile
from datetime import datetime
from django.conf import settings
from django.core.management import CommandParser
from jutil.command import SafeCommand
from jbank.models import WsEdiConnection


class Command(SafeCommand):
    help = 'Export WS-EDI connection'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('ws', type=int)
        parser.add_argument('--file', type=str)

    def do(self, *args, **options):
        ws = WsEdiConnection.objects.all().get(id=options['ws'])
        assert isinstance(ws, WsEdiConnection)

        filename = 'ws{}.zip'.format(ws.id)
        if options['file']:
            filename = options['file']

        files = []
        ws_data = {}
        for k, v in ws.__dict__.items():
            if not k.startswith('_') and k != 'id':
                if isinstance(v, (datetime, datetime)):
                    v = v.isoformat()
                ws_data[k] = v
                if k.endswith('_file'):
                    files.append(os.path.join(settings.MEDIA_ROOT, v))

        zf = zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED)
        zf.writestr('ws.json', json.dumps(ws_data))
        for file in files:
            zf.write(file, os.path.basename(file))
        zf.close()
        print('{} written'.format(filename))
